import asyncio
import datetime
import functools
import logging
import os
import re
from dataclasses import dataclass
from typing import List, Optional

import aiohttp
import bs4
import pandas as pd
import requests
from pytse_client import (
    config,
    symbols_data,
    translations,
    tse_settings,
    utils,
)
from pytse_client.download import download, download_ticker_client_types_record
from pytse_client.proxy.dto import ShareholderData
from pytse_client.proxy.tsetmc import get_day_shareholders_history
from pytse_client.ticker.api_extractors import (
    Order,
    TradeSummary,
    get_corporate_trade_summary,
    get_individual_trade_summary,
    get_orders,
)
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL
from tenacity import retry, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)
logger.addHandler(logging.NullHandler())


@dataclass
class RealtimeTickerInfo:
    last_price: Optional[float]
    adj_close: Optional[float]
    best_demand_vol: Optional[int]
    best_demand_price: Optional[float]
    best_supply_vol: Optional[int]
    best_supply_price: Optional[float]
    sell_orders: Optional[List[Order]]
    buy_orders: Optional[List[Order]]
    individual_trade_summary: Optional[TradeSummary]
    corporate_trade_summary: Optional[TradeSummary]


class Ticker:
    def __init__(
        self,
        symbol: str,
        index: Optional[str] = None,
        adjust: bool = False,
    ):
        self._index = index or symbols_data.get_ticker_index(symbol)
        self.symbol = symbol if index is None else self._index
        self.adjust = adjust
        self.daily_records_csv_path = (
            f"{config.DATA_BASE_PATH}/{self.symbol}.csv"
        )
        self.adjusted_daily_records_csv_path = (
            f"{config.DATA_BASE_PATH}/{self.symbol}-ت.csv"
        )
        self._url = tse_settings.TSE_TICKER_ADDRESS.format(self._index)
        self._info_url = tse_settings.TSE_ISNT_INFO_URL.format(self._index)
        self._client_types_url = TSE_CLIENT_TYPE_DATA_URL.format(self._index)
        self._history: pd.DataFrame = pd.DataFrame()

        if self.adjust:
            if os.path.exists(self.adjusted_daily_records_csv_path):
                self.from_file()
            else:
                self.from_web()
        else:
            if os.path.exists(self.daily_records_csv_path):
                self.from_file()
            else:
                self.from_web()

    def from_web(self):
        self._history = download(
            symbols=self.symbol,
            adjust=self.adjust,
        )[self.symbol]

    def from_file(self):
        if self.adjust:
            self._history = pd.read_csv(self.adjusted_daily_records_csv_path)
        else:
            self._history = pd.read_csv(self.daily_records_csv_path)
        self._history["date"] = pd.to_datetime(self._history["date"])

    @property
    def history(self):
        return self._history

    @property
    def url(self):
        return self._url

    @property
    def index(self):
        return self._index

    @property
    def instrument_id(self):
        """
        instrument id of a ticker is unique and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"InstrumentID='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def ci_sin(self):
        """
        instrument id of a ticker is like instrument_id and used for calling
        some apis from tsetmc
        """
        return re.findall(
            r"CIsin='([\w\d]*)|',$", self._ticker_page_response.text
        )[0]

    @property
    def title(self) -> str:
        return re.findall(r"Title='(.*?)',", self._ticker_page_response.text
                          )[0].split("-")[0].strip()

    @property
    def group_name(self) -> str:
        return re.findall(
            r"LSecVal='([\D]*)',", self._ticker_page_response.text
        )[0]

    @property
    def p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have p/e
        """
        adj_close = self.get_ticker_real_time_info_response().adj_close
        eps = self.eps
        if adj_close is None or eps is None or eps == 0:
            return None
        return self.get_ticker_real_time_info_response().adj_close / self.eps

    @property
    def group_p_e_ratio(self) -> Optional[float]:
        """
        Notes on usage: tickers like وملت does not have group P/E (gpe)
        """
        gpe = re.findall(
            r"SectorPE='([\d.]*)',", self._ticker_page_response.text
        )
        if not gpe or not gpe[0]:
            return None
        return float(gpe[0])

    @property
    def eps(self) -> Optional[float]:
        """
        Notes on usage: tickers like آسام does not have eps
        """
        eps = re.findall(
            r"EstimatedEPS='([-,\d]*)',", self._ticker_page_response.text
        )[0]
        if eps == "":
            return None
        return float(eps)

    @property
    def total_shares(self) -> float:
        return float(
            re.findall(r"ZTitad=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def base_volume(self) -> float:
        return float(
            re.findall(r"BaseVol=([-,\d]*),",
                       self._ticker_page_response.text)[0]
        )

    @property
    def client_types(self):
        return download_ticker_client_types_record(self._index)

    @property
    def trade_dates(self):
        return self._history["date"].to_list()

    @property
    def shareholders(self) -> pd.DataFrame:
        session = utils.requests_retry_session()
        page = session.get(self._shareholders_url, timeout=5)
        session.close()
        soup = bs4.BeautifulSoup(page.content, 'html.parser')
        table: bs4.PageElement = soup.find_all("table")[0]
        shareholders_df = utils.get_shareholders_html_table_as_csv(table)
        shareholders_df = shareholders_df.rename(
            columns=translations.SHAREHOLDERS_FIELD_MAPPINGS
        )
        return shareholders_df

    def get_shareholders_history(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None
    ) -> pd.DataFrame:
        """
            a helper function to use shareholders_history_async
        """
        return asyncio.run(
            self.get_shareholders_history_async(
                from_when,
                to_when,
                only_trade_days,
                session,
            ),
        )

    async def get_shareholders_history_async(
        self,
        from_when=datetime.timedelta(days=90),
        to_when=datetime.datetime.now(),
        only_trade_days=True,
        session=None,
    ) -> pd.DataFrame:
        requested_dates = utils.datetime_range(to_when - from_when, to_when)
        session_created = False
        if not session:
            session_created = True
            conn = aiohttp.TCPConnector(limit=25)
            session = aiohttp.ClientSession(connector=conn)
        tasks = []
        filtered_dates = list(
            filter(
                lambda date: not only_trade_days or
                (only_trade_days and date.date() in self.trade_dates),
                requested_dates,
            )
        )
        for date in filtered_dates:
            tasks.append(
                get_day_shareholders_history(self._index, date, session)
            )
        all_tickers_shareholders: List[List[ShareholderData]
                                       ] = await asyncio.gather(*tasks)
        if session_created is True:
            await session.close()

        rows = []
        for ticker_shareholders, info_date in zip(
            all_tickers_shareholders, filtered_dates
        ):
            for shareholder_data in ticker_shareholders:
                rows.append(
                    [
                        info_date,
                        shareholder_data.id,
                        shareholder_data.shares,
                        shareholder_data.percentage,
                        shareholder_data.instrument_id,
                        shareholder_data.name,
                        shareholder_data.change,
                    ]
                )
        return pd.DataFrame(
            data=rows,
            columns=[
                'date',
                'shareholder_id',
                'shareholder_shares',
                'shareholder_percentage',
                'shareholder_instrument_id',
                'shareholder_name',
                'change',
            ]
        )

    @property
    def last_price(self):
        return self.get_ticker_real_time_info_response().last_price

    @property
    def adj_close(self):
        return self.get_ticker_real_time_info_response().adj_close

    @property
    def best_demand_vol(self):
        return self.get_ticker_real_time_info_response().best_demand_vol

    @property
    def best_demand_price(self):
        return self.get_ticker_real_time_info_response().best_demand_price

    @property
    def best_supply_vol(self):
        return self.get_ticker_real_time_info_response().best_supply_vol

    @property
    def best_supply_price(self):
        return self.get_ticker_real_time_info_response().best_supply_price

    def get_ticker_real_time_info_response(self) -> RealtimeTickerInfo:
        """
        notes on usage:
        - Real time data might not be always available
        check for None values before usage
        """
        session = utils.requests_retry_session()
        response = session.get(self._info_url, timeout=5)
        session.close()

        # in some cases last price or adj price is undefined
        try:
            last_price = int(response.text.split()[1].split(",")[1])
        # when instead of number value is `F`
        except (ValueError, IndexError):
            last_price = None
        try:
            adj_close = int(response.text.split()[1].split(",")[2])
        except (ValueError, IndexError):
            adj_close = None

        response_sections_list = response.text.split(";")

        try:
            orders_section = response_sections_list[2]
            buy_orders, sell_orders = get_orders(orders_section)
            best_demand_vol = (
                buy_orders[0].volume if 0 < len(buy_orders) else None
            )
            best_demand_price = (
                buy_orders[0].price if 0 < len(buy_orders) else None
            )
            best_supply_vol = (
                sell_orders[0].volume if 0 < len(sell_orders) else None
            )
            best_supply_price = (
                sell_orders[0].price if 0 < len(sell_orders) else None
            )
        except (IndexError):
            buy_orders = []
            sell_orders = []
            logger.warning(
                f"""not enough sections in response to demand
                and supply information resp: {response_sections_list}"""
            )
            best_demand_vol = None
            best_demand_price = None
            best_supply_vol = None
            best_supply_price = None
            buy_orders = None
            sell_orders = None

        if len(response_sections_list) >= 4:
            trade_summary_section = response_sections_list[4]
            individual_trade_summary = get_individual_trade_summary(
                trade_summary_section
            )
            corporate_trade_summary = get_corporate_trade_summary(
                trade_summary_section
            )
        else:
            logger.warning(
                f"""not enough sections in response to extract trade summaries
                resp: {response_sections_list}"""
            )
            individual_trade_summary = None
            corporate_trade_summary = None

        return RealtimeTickerInfo(
            last_price,
            adj_close,
            best_demand_vol=best_demand_vol,
            best_demand_price=best_demand_price,
            best_supply_vol=best_supply_vol,
            best_supply_price=best_supply_price,
            buy_orders=buy_orders,
            sell_orders=sell_orders,
            individual_trade_summary=individual_trade_summary,
            corporate_trade_summary=corporate_trade_summary,
        )

    @property
    @functools.lru_cache()
    def _ticker_page_response(self):
        return utils.requests_retry_session().get(self._url, timeout=10)

    @functools.lru_cache()
    @retry(
        wait=wait_random(min=3, max=5),
        before_sleep=before_sleep_log(logger, logging.ERROR)
    )
    async def _get_ticker_daily_info_page_response(
        self, session, date
    ) -> requests.Response:
        async with session.get(
            tse_settings.INSTRUMENT_DAY_INFO_URL.format(
                index=self.index, date=date
            ),
        ) as response:
            response.raise_for_status()
            page = await response.text()
            logger.info(f"fetched date {date}")
            return page

    @property
    def _shareholders_url(self) -> str:
        return tse_settings.TSE_SHAREHOLDERS_URL.format(self.ci_sin)



import logging
from concurrent import futures
from io import StringIO
from pathlib import Path
from typing import Dict, List, Optional, Union

import jdatetime
import pandas as pd
from pytse_client import config, symbols_data, translations, tse_settings
from pytse_client.tse_settings import TSE_CLIENT_TYPE_DATA_URL
from pytse_client.utils import persian, requests_retry_session
from pytse_client.scraper.symbol_scraper import MarketSymbol
from requests import HTTPError
from requests.sessions import Session
from tenacity import retry, retry_if_exception_type, wait_random
from tenacity.before_sleep import before_sleep_log

logger = logging.getLogger(config.LOGGER_NAME)


def download(
    symbols: Union[List, str],
    write_to_csv: bool = False,
    include_jdate: bool = False,
    base_path: str = config.DATA_BASE_PATH,
    adjust: bool = False
) -> Dict[str, pd.DataFrame]:
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        session = requests_retry_session()
        for symbol in symbols:
            if symbol.isnumeric():
                ticker_indexes = [symbol]
            else:
                ticker_index = _handle_ticker_index(symbol)
                if ticker_index is None:
                    raise Exception(f"Cannot find symbol: {symbol}")
                ticker_indexes = symbols_data.get_ticker_old_index(symbol)
                ticker_indexes.insert(0, ticker_index)
            for index in ticker_indexes:
                future = executor.submit(
                    download_ticker_daily_record, index, session
                )
                future_to_symbol[future] = symbol

        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            try:
                df: pd.DataFrame = future.result()
            except pd.errors.EmptyDataError as ex:
                logger.error(
                    f"Cannot read daily trade records for symbol: {symbol}",
                    extra={"Error": ex}
                )
                continue

            df = df.iloc[::-1].reset_index(drop=True)
            df = df.rename(columns=translations.HISTORY_FIELD_MAPPINGS)
            df = df.drop(columns=["<PER>", "<TICKER>"])
            _adjust_data_frame(df, include_jdate)

            if symbol in df_list:
                df_list[symbol] = (
                    df_list[symbol].append(
                        df,
                        ignore_index=True,
                        sort=False
                    ).sort_values('date').reset_index(drop=True)
                )
            else:
                df_list[symbol] = df

            if adjust:
                df_list[symbol] = adjust_price(df_list[symbol])

            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                if adjust:
                    df_list[symbol].to_csv(
                        f'{base_path}/{symbol}-ت.csv',
                        index=False
                    )
                else:
                    df_list[symbol].to_csv(
                        f'{base_path}/{symbol}.csv',
                        index=False
                    )

    if len(df_list) != len(symbols):
        print("Warning, download did not complete, re-run the code")
    session.close()
    return df_list


def adjust_price(
    df: pd.DataFrame
) -> pd.DataFrame:
    """
    Adjust historical records of stock

    There is a capital increase/profit sharing,
    if today "Final Close Price" is not equal to next day
    "Yesterday Final Close Price" by using this ratio,
    performance adjustment of stocks is achieved

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with historical records.

    Returns
    -------
    pd.DataFrame
        DataFrame with adjusted historical records.

    Notes
    -----
    DataFrame can not be empty or else it makes runtime error
    Type of DataFrame must be RangeIndex to make proper range of records
    that need to be modified

    diff: list
        list of indexs of the day after capital increase/profit sharing
    ratio_list: List
        List of ratios to adjust historical data of stock
    ratio: Float
        ratio = df.loc[i].adjClose / df.loc[i+1].yesterday

    Description
    -----------
    #Note: adjustment does not include Tenth and twentieth days
    df.index = range(0,101,1)
    #step is 1
    step = df.index.step
    diff = [10,20]
    ratio_list = [0.5, 0.8]
    df.loc[0:10-step, [open,...]] * ratio[0]
    df.loc[10:20-step, [open,...]] * ratio[1]
    """
    if df.empty or not isinstance(df.index, pd.core.indexes.range.RangeIndex):
        return df

    new_df = df.copy()
    step = new_df.index.step
    diff = list(new_df.index[new_df.shift(1).adjClose != new_df.yesterday])
    if len(diff) > 0:
        diff.pop(0)
    ratio = 1
    ratio_list = []
    for i in diff[::-1]:
        ratio *= (
            new_df.loc[i, 'yesterday'] / new_df.shift(1).loc[i, 'adjClose']
        )
        ratio_list.insert(0, ratio)
    for i, k in enumerate(diff):
        if i == 0:
            start = new_df.index.start
        else:
            start = diff[i-1]
        end = diff[i]-step
        new_df.loc[start:end, [
            'open',
            'high',
            'low',
            'close',
            'adjClose',
            'yesterday',
        ]] = round(
            new_df.loc[start:end, [
                'open',
                'high',
                'low',
                'close',
                'adjClose',
                'yesterday',
            ]] * ratio_list[i]
        )

    return new_df


def _adjust_data_frame(df, include_jdate):
    df.date = pd.to_datetime(df.date, format="%Y%m%d")
    if include_jdate:
        df['jdate'] = ""
        df.jdate = df.date.apply(
            lambda gregorian: jdatetime.date.fromgregorian(date=gregorian)
        )


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_random(min=1, max=4),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def download_ticker_daily_record(ticker_index: str, session: Session):
    url = tse_settings.TSE_TICKER_EXPORT_DATA_ADDRESS.format(ticker_index)
    response = session.get(url, timeout=10)
    if 400 <= response.status_code < 500:
        logger.error(
            f"Cannot read daily trade records from the url: {url}",
            extra={
                "response": response.text,
                "code": response.status_code
            }
        )

    response.raise_for_status()

    data = StringIO(response.text)
    return pd.read_csv(data)


def download_client_types_records(
    symbols: Union[List, str],
    write_to_csv: bool = False,
    include_jdate: bool = False,
    base_path: str = config.CLIENT_TYPES_DATA_BASE_PATH
):
    if symbols == "all":
        symbols = symbols_data.all_symbols()
    elif isinstance(symbols, str):
        symbols = [symbols]

    df_list = {}
    future_to_symbol = {}
    with futures.ThreadPoolExecutor(max_workers=10) as executor:
        for symbol in symbols:
            ticker_index = _handle_ticker_index(symbol)
            if ticker_index is None:
                raise Exception(f"Cannot find symbol: {symbol}")
            future = executor.submit(
                download_ticker_client_types_record, ticker_index
            )
            future_to_symbol[future] = symbol
        for future in futures.as_completed(future_to_symbol):
            symbol = future_to_symbol[future]
            df: pd.DataFrame = future.result()
            # ignore failures
            if df is None:
                continue
            _adjust_data_frame(df, include_jdate)
            df_list[symbol] = df
            if write_to_csv:
                Path(base_path).mkdir(parents=True, exist_ok=True)
                df.to_csv(f'{base_path}/{symbol}.csv')

    if len(df_list) != len(symbols):
        print(
            """could not download client types for all the symbols make
        sure you have what you need or re-run the function"""
        )
    return df_list


def _handle_ticker_index(symbol):
    ticker_index = symbols_data.get_ticker_index(symbol)
    if ticker_index is None:
        market_symbol = get_symbol_info(symbol)
        if market_symbol is not None:
            symbols_data.append_symbol_to_file(market_symbol)
            ticker_index = market_symbol.index
    return ticker_index


@retry(
    retry=retry_if_exception_type(HTTPError),
    wait=wait_random(min=1, max=4),
    before_sleep=before_sleep_log(logger, logging.DEBUG)
)
def download_ticker_client_types_record(ticker_index: Optional[str]):
    data = _extract_ticker_client_types_data(ticker_index)
    if len(data) == 1:
        logger.warning(
            f"""Cannot create client types data for ticker{ticker_index}
             from data: {data}""",
            extra={"ticker_index": ticker_index}
        )
        return None
    client_types_data_frame = pd.DataFrame(
        data,
        columns=[
            "date", "individual_buy_count", "corporate_buy_count",
            "individual_sell_count", "corporate_sell_count",
            "individual_buy_vol", "corporate_buy_vol", "individual_sell_vol",
            "corporate_sell_vol", "individual_buy_value",
            "corporate_buy_value", "individual_sell_value",
            "corporate_sell_value"
        ]
    )
    for i in [
        "individual_buy_", "individual_sell_", "corporate_buy_",
        "corporate_sell_"
    ]:
        client_types_data_frame[f"{i}mean_price"] = (
            client_types_data_frame[f"{i}value"].astype(float) /
            client_types_data_frame[f"{i}vol"].astype(float)
        )
    client_types_data_frame["individual_ownership_change"] = (
        client_types_data_frame["corporate_sell_vol"].astype(float) -
        client_types_data_frame["corporate_buy_vol"].astype(float)
    )
    return client_types_data_frame


def _extract_ticker_client_types_data(ticker_index: str) -> List:
    url = TSE_CLIENT_TYPE_DATA_URL.format(ticker_index)
    with requests_retry_session() as session:
        response = session.get(url, timeout=5)
    data = response.text.split(";")
    data = [row.split(",") for row in data]
    return data


def get_symbol_id(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception("Sorry, tse server did not respond")

    symbol_full_info = response.text.split(';')[0].split(',')
    if persian.replace_arabic(symbol_name) == symbol_full_info[0].strip():
        return symbol_full_info[2]  # symbol id
    return None


def get_symbol_info(symbol_name: str):
    url = tse_settings.TSE_SYMBOL_ID_URL.format(symbol_name.strip())
    response = requests_retry_session().get(url, timeout=10)
    try:
        response.raise_for_status()
    except HTTPError:
        raise Exception(f"{symbol_name}: Sorry, tse server did not respond")

    symbols = response.text.split(';')
    market_symbol = MarketSymbol(
                code=None,
                symbol=None,
                index=None,
                name=None,
                old=[],
            )
    for symbol_full_info in symbols:
        if symbol_full_info.strip() == "":
            continue
        symbol_full_info = symbol_full_info.split(',')
        if persian.replace_arabic(symbol_full_info[0]) == symbol_name:
            # if symbol id is active
            if symbol_full_info[7] == '1':
                market_symbol.symbol = persian.replace_arabic(
                    symbol_full_info[0]
                )
                market_symbol.name = persian.replace_arabic(
                    symbol_full_info[1]
                )
                market_symbol.index = symbol_full_info[2]  # active symbol id
            else:
                market_symbol.old.append(symbol_full_info[2])  # old symbol id

    if market_symbol.index is None:
        return None
    return market_symbol



"""
This example is about getting data for a ticker.
use this example if you want to get data for one ticker
"""

from pytse_client import Ticker, download
import pandas as pd

# to be able to see whole DataFrame columns
pd.set_option('display.max_columns', 20)

download(symbols="وبملت", write_to_csv=True)  # optional
ticker = Ticker("وبملت")
print(ticker.history)  # سابقه قیمت سهم
print(ticker.client_types)  # حقیقی حقوقی
print(ticker.title)  # نام شرکت
print(ticker.url)  # آدرس صفحه سهم
print(ticker.group_name)  # نام گروه
print(ticker.eps)  # eps
print(ticker.p_e_ratio)  # P/E
print(ticker.group_p_e_ratio)  # group P/E
print(ticker.base_volume)  # حجم مبنا
print(ticker.last_price)  # آخرین معامله
print(ticker.adj_close)  # قیمت پایانی
print(ticker.shareholders)  # اطلاعات سهام داران عمده
print(ticker.shareholders.percentage.sum())  # جمع سهام داران
print(ticker.total_shares)

real_time_data = ticker.get_ticker_real_time_info_response()
print(real_time_data.individual_trade_summary.buy_count)
print(real_time_data.individual_trade_summary.buy_vol)
print(real_time_data.individual_trade_summary.sell_count)
print(real_time_data.individual_trade_summary.sell_vol)
print(real_time_data.corporate_trade_summary.buy_count)
print(real_time_data.corporate_trade_summary.buy_vol)
print(real_time_data.corporate_trade_summary.sell_count)
print(real_time_data.corporate_trade_summary.sell_vol)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytse_client',
 'pytse_client.data',
 'pytse_client.examples',
 'pytse_client.proxy',
 'pytse_client.scraper',
 'pytse_client.scripts',
 'pytse_client.ticker',
 'pytse_client.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'jdatetime>=3.6.2,<4.0.0',
 'pandas',
 'requests>=2.23.0,<3.0.0',
 'tenacity>=7.0.0,<8.0.0']

setup_kwargs = {
    'name': 'pytse-client',
    'version': '0.10.0rc1',
    'description': 'tehran stock exchange(TSE) client in python',
    'long_description': '<div dir="rtl">\n\n# دریافت اطلاعات بازار بورس تهران\n\n![Python application](https://github.com/Glyphack/pytse-client/workflows/Python%20application/badge.svg)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/Glyphack/pytse-client.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/Glyphack/pytse-client/context:python)\n[![Discord Chat](https://img.shields.io/discord/730808323808559106?label=discord)](https://discord.gg/ampPDKHpVv)\n\nبا استفاده از pytse client میتونید به دیتای بازار بورس تهران در پایتون دسترسی داشته باشید.\nهدف حل مشکلات گرفتن اطلاعات بروز از سایت بازار بورس تهران هست.\n\n## میخواید مشارکت کنید؟\nلطفا [این صفحه](https://github.com/Glyphack/pytse-client/blob/master/CONTRIBUTING.md) رو مطالعه کنید\n\n- [دریافت اطلاعات بازار بورس تهران](#دریافت-اطلاعات-بازار-بورس-تهران)\n  - [میخواید مشارکت کنید؟](#میخواید-مشارکت-کنید)\n  - [قابلیت ها](#قابلیت-ها)\n  - [نصب](#نصب)\n  - [نحوه استفاده](#نحوه-استفاده)\n    - [دانلود سابقه سهم\u200cها](#دانلود-سابقه-سهمها)\n    - [دانلود سابقه معاملات حقیقی و حقوقی به صورت مجزا](#دانلود-سابقه-معاملات-حقیقی-و-حقوقی-به-صورت-مجزا)\n    - [ماژول Ticker](#ماژول-ticker)\n        - [۱نکته](#۱نکته)\n        - [نکته۲](#نکته۲)\n      - [اطلاعات نماد\u200cهای حذف شده](#اطلاعات-نمادهای-حذف-شده)\n      - [اطلاعات حقیقی و حقوقی](#اطلاعات-حقیقی-و-حقوقی)\n      - [سهامداران عمده](#سهامداران-عمده)\n        - [شناور سهم](#شناور-سهم)\n        - [تاریخچه\u200cی سهام داران عمده](#تاریخچهی-سهام-داران-عمده)\n        - [اطلاعات لحظه\u200cای سهام](#اطلاعات-لحظهای-سهام)\n      - [کامیونیتی](#کامیونیتی)\n      - [پکیج های مورد نیاز:](#پکیج-های-مورد-نیاز)\n      - [الهام گرفته از:](#الهام-گرفته-از)\n\n## قابلیت ها\n\n- دریافت اطلاعات روز های معاملاتی هر سهم و قابلیت ذخیره سازی\n- قابلیت گرفتن اطلاعات یک سهام مانند گروه سهام و اطلاعات معاملات حقیقی و حقوقی\n- دریافت اطلاعات فاندامنتال یک نماد شامل(eps, p/e ,حجم مبنا)\n\n## نصب\n\n<div dir="ltr">\n\n```bash\npip install pytse-client\n```\n\n</div>\n\n## نحوه استفاده\n\n### دانلود سابقه سهم\u200cها\n\nبا استفاده از این تابع میتوان سابقه سهام هارو دریافت کرد و هم اون رو ذخیره و هم توی کد استفاده کرد\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\ntickers = tse.download(symbols="all", write_to_csv=True)\ntickers["ولملت"] # history\n\n            date     open     high  ...     volume  count    close\n0     2009-02-18   1050.0   1050.0  ...  330851245    800   1050.0\n1     2009-02-21   1051.0   1076.0  ...  335334212   6457   1057.0\n2     2009-02-22   1065.0   1074.0  ...    8435464    603   1055.0\n3     2009-02-23   1066.0   1067.0  ...    8570222    937   1060.0\n4     2009-02-25   1061.0   1064.0  ...    7434309    616   1060.0\n...          ...      ...      ...  ...        ...    ...      ...\n2323  2020-04-14   9322.0   9551.0  ...  105551315  13536   9400.0\n2324  2020-04-15   9410.0   9815.0  ...  201457026  11322   9815.0\n2325  2020-04-18  10283.0  10283.0  ...  142377245   8929  10283.0\n2326  2020-04-19  10797.0  10797.0  ...  292985635  22208  10380.0\n2327  2020-04-20  10600.0  11268.0  ...  295590437  16313  11268.0\n```\n\n</div>\n\nبرای دانلود سابقه یک یا چند سهم کافی هست اسم اون ها به تابع داده بشه:\n\n<div dir="rtl">\n\nهمچنین با گذاشتن\n`write_to_csv=True`\nسابقه سهم توی فایلی با اسم سهم نوشته میشه\n\nسابقه سهم در قالب `Dataframe` است\n\nدرصورتی که می خواهید تاریخ شمسی به خروجی اضافه شود می توانید با گذاشتن\n`include_jdate=True`\nاین امکان را فراهم کنید\n\n</div>\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\ntse.download(symbols="وبملت", write_to_csv=True)\ntse.download(symbols="وبملت", write_to_csv=True, include_jdate=True)\ntse.download(symbols=["وبملت", "ولملت"], write_to_csv=True)\n```\n\n</div>\n\n### دانلود سابقه معاملات حقیقی و حقوقی به صورت مجزا\n\nبرای دانلود سابقه معاملات حقیقی و حقوقی برای تمامی نمادها میتوان از تابع زیر استفاده کرد\n\n<div dir="ltr">\n\n```python\nfrom pytse_client import download_client_types_records\n\nif __name__ == \'__main__\':\n\n  records_dict = download_client_types_records("all")\n  print(records_dict["فولاد"])\n  #Output\ndate         individual_buy_count  ... individual_ownership_change\n\n2020-09-01                36298  ...                   -691857.0\n2020-08-31                58185  ...                  83789408.0\n2020-08-26                  461  ...                  21647730.0\n2020-08-25                 1248  ...                  14716846.0\n2020-08-24                38291  ...                -238454702.0\n...                         ...  ...                         ...\n2008-12-02                    7  ...                    -10000.0\n2008-12-01                    8  ...                         0.0\n2008-11-30                   10  ...                    -12781.0\n2008-11-29                  116  ...                   4596856.0\n2008-11-26                   14  ...                    -20000.0\n\n[2518 rows x 17 columns]\n\n```\n\n</div>\n\nمشابه تابع قبلی میتوان نتایج را ذخیره کرد\n\n<div dir="ltr">\n\n```python\nfrom pytse_client import download_client_types_records\nif __name__ == \'__main__\':\n\n  #Records are saved as a .csv file with the same name of ticer\'s\n  records = download_client_types_records("فولاد", write_to_csv=True)\n```\n\n</div>\n\n### ماژول Ticker\n\nاین ماژول برای کار با دیتای یک سهم خاص هست و با گرفتن نماد اطلاعات موجود رو میده\n\nبرای مثال:\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\ntse.download(symbols="وبملت", write_to_csv=True)  # optional\nticker = tse.Ticker("وبملت")\nprint(ticker.history)  # سابقه قیمت سهم\nprint(ticker.client_types)  # حقیقی حقوقی\nprint(ticker.title)  # نام شرکت\nبانك ملت (وبملت)\nprint(ticker.url)  # آدرس صفحه سهم\nhttp://tsetmc.com/Loader.aspx?ParTree=151311&i=778253364357513\nprint(ticker.group_name)  # نام گروه\nبانكها و موسسات اعتباري\nprint(ticker.eps)  # eps\n2725.0\nprint(ticker.p_e_ratio)  # P/E\n6.1478899082568805\nprint(ticker.group_p_e_ratio)  # group P/E\n18.0\nprint(ticker.base_volume)  # حجم مبنا\n7322431.0\nprint(ticker.last_price)  # آخرین معامله\n17316\nprint(ticker.adj_close)  # قیمت پایانی\n16753\nprint(ticker.best_supply_price)  # قیمت بهترین تقاضا\n26700\nprint(ticker.best_supply_vol)  # حجم بهترین تقاضا\n357062\nprint(ticker.best_demand_price)  # قیمت بهترین عرضه\n26700\nprint(ticker.best_demand_vol)  # حجم بهترین عرضه\n576608\nprint(ticker.shareholders)  # اطلاعات سهام داران عمده\n\nprint(ticker.get_shareholders_history())) # تاریخچه\u200cی سهام داران عمده\nprint(ticker.get_ticker_real_time_info_response()) # اطلاعات لحظه\u200cای مانند قیمت و پیشنهادات خرید و فروش\n\n\n```\n\n</div>\n\nبرای استفاده لازم نیست حتما تابع دانلود صدا زده بشه.\nاگر این کد رو بدون دانلود کردن سهم استفاده کنید خودش اطلاعات سهم رو از سایت میگیره،\nاما اگر قبل از اون از دانلود استفاده کرده باشید\nبه جای گرفتن از اینترنت اطلاعات رو از روی فایل میخونه که سریع تر هست\n\n##### ۱نکته\n\nطبق تجربه\u200c ای که داشتم چون گاهی اوقات سایت بورس مدت زیادی طول میکشه تا اطلاعات رو بفرسته یا بعضی مواقع نمیفرسته بهتر هست که اول تابع دانلود رو استفاده کنید برای سهم\u200cهایی که لازم هست و بعد با دیتای اون ها کار کنید.\n\n##### نکته۲\n\nبعضی از ویژگی\u200cها برای همه\u200cی سهم ها در دسترس نیست. برای مثال بعضی از سهم ها دارای آخرین قیمت یا پی به ای یا ای پی اس نیستند. مقدار این ویژگی\u200cها در صورت نبودن برابر با `None` خواهد بود. پس باید در برنامه خود اینکه این مقادیر وجود دارند را بررسی کنید.\n\n#### اطلاعات نماد\u200cهای حذف شده\nتعدادی از نماد\u200cها توی سایت به شکل حذف شده هستند. برای گرفتن دیتای این نماد\u200cها از ماژول تیکر به شکل زیر استفاده کنید:\nبرای مثال در سهام \nhttp://www.tsetmc.com/Loader.aspx?ParTree=151311&i=25165947991415904\n\n<div dir="ltr">\n\n```python\nimport pytse_client as tse\n\nticker = tse.Ticker("", index="25165947991415904")\n```\n\n</div>\n\nمقدار  \u200d`index` را با مقدار جلوی `i=` جایگزین میکنیم.\n#### اطلاعات حقیقی و حقوقی\n\nاطلاعات خرید و فروش حقیقی و حقوقی سهام رو میشه از طریق `ticker.client_types` گرفت این اطلاعات یه DataFrame شامل اطلاعات موجود در تب حقیقی حقوقی(تب بنفشی که در این [صفحه](http://www.tsetmc.com/Loader.aspx?ParTree=151311&i=778253364357513) هست) سهم هست:\n\n<div dir="ltr">\n\n```\ndate : تاریخ\nindividual_buy_count : تعداد معاملات خرید حقیقی\ncorporate_buy_count : تعداد معلاملات خرید حقوقی\nindividual_sell_count : تعداد معاملات فروش حقیقی\ncorporate_sell_count : تعداد معلاملات فروش حقوقی\nindividual_buy_vol : حجم خرید حقیقی\ncorporate_buy_vol : حجم خرید حقوقی\nindividual_sell_vol : حجم فروش حقیقی\ncorporate_sell_value : حجم فروش حقوقی\nindividual_buy_mean_price : قیمت میانگین خرید حقیقی\nindividual_sell_mean_price : قیمت میانگین فروش حقیقی\ncorporate_buy_mean_price : قیمت میانگین خرید حقوقی\ncorporate_sell_mean_price : قیمت میانگین فروش حقوقی\nindividual_ownership_change : تغییر مالکیت حقوقی به حقیقی\n```\n\n</div>\n\n#### سهامداران عمده\n\nسهام داران عمده اطلاعات داخل این [صفحه](http://tsetmc.com/Loader.aspx?Partree=15131T&c=IRO1BMLT0007) هست.\nاین اطلاعات رو میشه با `shareholders` گرفت که یک DataFrame هست.\n\n\n<div dir="ltr">\n\n```\nticker = Ticker("وبملت")\nprint(ticker.shareholders) # اطلاعات سهام داران عمده\n\n change   percentage       share                                 shareholder  \n0   دولت جمهوري اسلامي ايران                    23,114,768,760  11.160     0     \n1   صندوق تامين آتيه كاركنان بانك ملت           13,353,035,330  6.440      0      \n2   صندوق سرمايه گذاري واسطه گري مالي يكم       11,748,764,647  5.670      0      \n3   شركت پتروشيمي فن آوران-سهامي عام-           9,253,327,080   4.460      0      \n4   شركت گروه مالي ملت-سهام عام-                8,933,698,834   4.310      0      \n5   صندوق سرمايه گذاري.ا.بازارگرداني ملت     8,395,500,914   4.050      0   \n6   شركت سرمايه گذاري صباتامين-سهامي عام-       7,659,597,269   3.690      0      \n7   شركت تعاوني معين آتيه خواهان                4,561,801,327   2.200      0      \n8   شركت س اتهران س.خ-م ك م ف ع-                4,278,903,677   2.060      0      \n9   شركت گروه توسعه مالي مهرآيندگان-سهامي عام-  4,161,561,525   2.000      0      \n10  شركت س اخراسان رضوي س.خ-م ك م ف ع-          3,442,236,423   1.660      0      \n11  شركت س افارس س.خ-م ك م ف ع-                 2,593,956,288   1.250      0      \n12  شركت س اخوزستان س.خ-م ك م ف ع-              2,526,080,803   1.220      0      \n13  شركت شيرين عسل-سهامي خاص-                   2,496,936,881   1.200      0      \n14  شركت سرمايه گذاري ملي ايران-سهامي عام-      2,423,674,676   1.170      0      \n15  شركت س ااصفهان س.خ-م ك م ف ع-               2,274,221,331   1.090      0      \n```\n</div>\n\n##### شناور سهم\nبرای مثال میشه با استفاده از سهامداران عمده شناور سهم رو حساب کرد:\n\n<div dir="ltr">\n\n```\nticker.shareholders.percentage.sum() # جمع سهام داران\n53.63\n\n100 - ticker.shareholders.percentage.sum() # شناور سهم \n46.37\n```\n</div>\n\n##### تاریخچه\u200cی سهام داران عمده\nبا استفاده از تابع get_shareholders_history میشه اطلاعات سهام داران عمده رو گرفت\nورودی\u200cهای تابع:\n```python\nfrom_when=datetime.timedelta(days=90), تعداد روز\u200cهای گذشته پیشفرض ۹۰ روز گذشته است\nto_when=datetime.datetime.now(), تا چه تاریخی اطلاعات گرفته شود پیشفرض امروز هست\nonly_trade_days=True, فقط روز\u200cهای معاملاتی پیش فرض بله\n```\nخروجی این تابع یک دیتا فریم به شکل زیر هست که به صورت csv \n<div dir="ltr">\n\n```\n,date,shareholder_id,shareholder_shares,shareholder_percentage,shareholder_instrument_id,shareholder_name,change\n0,2021-08-30 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n1,2021-08-30 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n2,2021-08-30 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n3,2021-08-30 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n4,2021-08-30 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n5,2021-08-31 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n6,2021-08-31 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n7,2021-08-31 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n8,2021-08-31 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n9,2021-08-31 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n10,2021-09-01 17:01:23.037957,273,2910355428.0,32.19,IRO1RSAP0000,شركت ايراني توليداتومبيل-سايپا-,1\n11,2021-09-01 17:01:23.037957,406,975144471.0,10.78,IRO1RSAP0000,شركت سايپا,1\n12,2021-09-01 17:01:23.037957,50264,454000000.0,5.02,IRO1RSAP0000,شركت ايراني توليداتومبيل سايپا-سهامي عام-,1\n13,2021-09-01 17:01:23.037957,42636,409843922.0,4.53,IRO1RSAP0000,شركت سرمايه گذاري وتوسعه صنعتي نيوان ابتكارس.ع,1\n14,2021-09-01 17:01:23.037957,46966,116002189.0,1.28,IRO1RSAP0000,BFMصندوق.س.ا.بازارگرداني سهم آشنايكم,1\n\n```\n</div>\n\n<div id="qa" />\nگرفتن این دیتا کار زمان بری هست(با توجه به تعداد روزی که لازم دارید) و سریع کردن کار با کد به راحتی امکان پذیر نیست. سعی نکنید با همزمان اجرا کردن این تابع برای سهم\u200cهای مختلف روند رو سریع\u200cتر کنید چون سایت ip رو بلاک میکنه.\nاگر موقع اجرای کد پیغام زیر را به تعداد زیاد گرفتید (مثلا هر ثانیه این پیغام اومد) یعنی آیپی شما توسط سایت بورس بلاک شده و چند دقیقه صبر کنید و دوباره ادامه بدید.\n<div dir="ltr">\n\n```\nRetrying pytse_client.ticker.ticker.Ticker._get_ticker_daily_info_page_response in 1.3127419515957892 seconds as it raised ClientResponseError: 500, message=\'Internal Server Error\', url=URL(\'http://cdn.tsetmc.com/Loader.aspx?ParTree=15131P&i=56574323121551263&d=20210220\').\n```\n</div>\n\n##### اطلاعات لحظه\u200cای سهام\nاز طریق تابع `get_ticker_real_time_info_response` میشه اطلاعات لحظه\u200cای سهام رو گرفت.\nنمونه\u200cی استفاده\n\n<div dir="ltr">\n\n```python\nreal_time_data = ticker.get_ticker_real_time_info_response()\n\nprint(real_time_data.buy_orders) # پیشنهادات خرید\nprint(real_time_data.sell_orders) # پیشنهادات فروش\nprint(real_time_data.best_supply_price)  # قیمت بهترین تقاضا\nprint(real_time_data.best_supply_vol)  # حجم بهترین تقاضا\nprint(real_time_data.best_demand_price)  # قیمت بهترین عرضه\nprint(real_time_data.best_demand_vol)  # حجم بهترین عرضه\nprint(real_time_data.adj_close) # آخرین معامله\nprint(real_time_data.last_price) # قیمت پایانی\n\n# پیشنهادات فروش\nfor sell_order in real_time_data.sell_orders:\n    print(sell_order.volume, sell_order.count, sell_order.price)\n\n# پشنهادات فروش\nfor buy_order in real_time_data.buy_orders:\n    print(buy_order.volume, buy_order.count, buy_order.price)\n\n# اطلاعات خرید و فروش حقیقی و حقوقی\nprint(real_time_data.individual_trade_summary.buy_count)\nprint(real_time_data.individual_trade_summary.buy_vol)\nprint(real_time_data.individual_trade_summary.sell_count)\nprint(real_time_data.individual_trade_summary.sell_vol)\nprint(real_time_data.corporate_trade_summary.buy_count)\nprint(real_time_data.corporate_trade_summary.buy_vol)\nprint(real_time_data.corporate_trade_summary.sell_count)\nprint(real_time_data.corporate_trade_summary.sell_vol)\n```\n</div>\n\n#### کامیونیتی\n\nاگر درباره پکیج یا استفاده از اون سوالی دارید میتونید توی سرور دیسکورد بپرسید.\n\nhttps://discord.gg/ampPDKHpVv\n\n<div id="required-packages" />\n\n#### پکیج های مورد نیاز:\n\n- [Pandas](https://github.com/pydata/pandas)\n- [Requests](http://docs.python-requests.org/en/master/)\n- [jdatetime](https://github.com/slashmili/python-jalali)\n- [beautifulsoup4](https://pypi.org/project/beautifulsoup4)\n\n<div id="credits" />\n\n#### الهام گرفته از:\n\n- [tehran_stocks](https://github.com/ghodsizadeh/tehran-stocks)\n- [yfinance](https://github.com/ranaroussi/yfinance)\n\n</div>\n',
    'author': 'glyphack',
    'author_email': 'sh.hooshyari@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
